from pydantic import BaseModel, Field
from ._scheme import SchemeConfig
from ._installer import InstallerConfig

CONFIG_FILE_NAME = "ldm.yml"


class DependencyConfig(BaseModel):
    schemes: dict[str, SchemeConfig] | None = Field(default_factory=lambda: {})
    dependencies: dict[str, str]
    config: InstallerConfig | None = Field(default_factory=lambda: InstallerConfig())



from pydantic import BaseModel, Field, AliasChoices


class SchemeConfig(BaseModel):
    uses: str
    with_: dict | None = Field(
        default=None,
        validation_alias=AliasChoices("with_", "with"),
    )



import abc
from ..component import Component
from ..logger import Logger, ClickLogger


class Command(Component, metaclass=abc.ABCMeta):
    def __init__(
        self,
        *,
        logger: Logger = ClickLogger(),
    ):
        super().__init__(logger=logger)

    @abc.abstractmethod
    def run(self, *args, **kwargs):
        raise NotImplementedError



import httpx
from os import makedirs
from os.path import exists, dirname
from ldm.logger import Logger
from ..parse import parse_entry
from ._scheme import Scheme


class HTTPScheme(Scheme):
    def __init__(
        self,
        name: str,
        encoding: str | None = None,
        headers: dict | None = None,
        cookies: dict | None = None,
        *,
        logger: Logger | None = None,
    ) -> None:
        super().__init__(name, logger=logger)
        self.encoding = encoding or "utf-8"
        self.headers = headers
        self.cookies = cookies

    def install(self, entry: str) -> None:
        # Parse the entry
        source, destination = parse_entry(entry)

        content = httpx.get(
            source,
            headers=self.headers,
            cookies=self.cookies,
        ).content.decode(self.encoding)

        if not exists(dirname(destination)):
            makedirs(dirname(destination), exist_ok=True)

        with open(destination, "w") as f:
            f.write(content)



from typing import Callable
from traceback import format_exc
from ..logger import Logger


def error_boundary(func: Callable, logger: Logger):
    try:
        func()
    except Exception as e:
        for line in format_exc().split("\n"):
            logger.debug(line)
        for line in str(e).split("\n"):
            logger.error(line)
        exit(1)



from ...config import SequentialConfig
from ._strategy import InstallStrategy


class SequentialInstallStrategy(InstallStrategy[SequentialConfig]):
    def install(self, dependencies):
        self.logger.debug("Installing dependencies sequentially")
        for dependency in dependencies:
            self.logger.debug(f"Installing: {dependency.name}")
            dependency.install()
            self.logger.info(f"Installed: {dependency.name}")



import click
from ..commands import InstallCommand
from ._logger import create_logger
from ._error import error_boundary


@click.group()
def cli():
    pass


@cli.command("install")
@click.argument("dependencies", nargs=-1)
@click.option("--debug", is_flag=True, default=False)
def install(dependencies: list[str], debug: bool):
    logger = create_logger(debug)

    command = InstallCommand(logger=logger)
    error_boundary(
        lambda: command.run(targets=dependencies or None),
        logger,
    )



import abc
from ...component import Component
from ..scheme import Scheme


class SchemeFactory(Component, metaclass=abc.ABCMeta):
    name: str

    @abc.abstractmethod
    def create(self, config: dict) -> Scheme:
        raise NotImplementedError



import sys
import click
from ._logger import Logger


class ClickLogger(Logger):
    def debug(self, message: str, *args, **kwargs) -> None:
        if self.level > 1:
            return
        kwargs.setdefault("fg", "blue")
        message = f"[ldm] |   DEBUG | {message}"
        click.echo(
            click.style(message, *args, **kwargs),
            file=sys.stdout,
        )

    def info(self, message: str, *args, **kwargs) -> None:
        if self.level > 2:
            return
        message = f"[ldm] |    INFO | {message}"
        click.echo(
            click.style(message, *args, **kwargs),
            file=sys.stdout,
        )

    def success(self, message: str, *args, **kwargs) -> None:
        if self.level > 3:
            return
        message = f"[ldm] | SUCCESS | {message}"
        kwargs.setdefault("fg", "green")
        click.echo(
            click.style(message, *args, **kwargs),
            file=sys.stdout,
        )

    def warning(self, message: str, *args, **kwargs) -> None:
        if self.level > 4:
            return
        message = f"[ldm] | WARNING | {message}"
        kwargs.setdefault("fg", "yellow")
        click.echo(
            click.style(message, *args, **kwargs),
            file=sys.stdout,
        )

    def error(self, message: str, *args, **kwargs) -> None:
        if self.level > 5:
            return
        message = f"[ldm] |   ERROR | {message}"
        kwargs.setdefault("fg", "red")
        click.echo(
            click.style(message, *args, **kwargs),
            file=sys.stderr,
        )



import abc
from typing import Generic, TypeVar
from ldm.logger import Logger
from ....component import Component
from .._dependency import Dependency

Config = TypeVar("Config")


class InstallStrategy(Component, Generic[Config], metaclass=abc.ABCMeta):
    def __init__(
        self,
        config: Config,
        *,
        logger: Logger | None = None,
    ):
        super().__init__(logger=logger)
        self.config = config

    @abc.abstractmethod
    def install(self, dependencies: list[Dependency]):
        raise NotImplementedError



from typing import Callable
from ldm.logger import Logger
from ...component import Component


class Dependency(Component):
    def __init__(
        self,
        name: str,
        install: Callable[[], None],
        *,
        logger: Logger | None = None,
    ) -> None:
        super().__init__(logger)
        self.name = name
        self.install = install



import os
from os import getcwd
from os.path import join, exists
from yaml_replace import YAMLTemplate
from ._command import Command
from ..api.installer import Installer
from ..api.config import CONFIG_FILE_NAME, DependencyConfig
from dotenv import load_dotenv


class InstallCommand(Command):
    def run(self, targets: list[str] | None = None):
        # Read configuration file
        config = read_config()
        if config is None:
            self.logger.error("No configuration file found")
            exit(1)

        self.logger.debug("Read configuration file: ")
        for line in config.model_dump_json(indent=2).split("\n"):
            self.logger.debug(line)

        Installer(logger=self.logger).install(config, targets)
        exit(0)


def read_config() -> DependencyConfig | None:
    if exists(CONFIG_FILE_NAME):
        with open(CONFIG_FILE_NAME, "r") as f:
            # Read configuration file
            template = YAMLTemplate(f.read())
            config = DependencyConfig(**template.render({}))

            # Load environment variables
            for env_file in config.config.environment.env_files:
                load_dotenv(dotenv_path=join(getcwd(), env_file))

            # Rerender configuration file with environment variables
            return DependencyConfig(**template.render(os.environ))
    return None



from pydantic import BaseModel, model_validator, Field, AliasChoices


class SequentialConfig(BaseModel):
    pass


class ParallelConfig(BaseModel):
    workers: int = 8


class EnvironmentConfig(BaseModel):
    from_: str | list[str] | None = Field(
        default=".env",
        validation_alias=AliasChoices("from_", "from"),
    )

    @property
    def env_files(self) -> list[str]:
        if self.from_ is None:
            return []
        if isinstance(self.from_, str):
            return [self.from_]
        return self.from_


class InstallerConfig(BaseModel):
    sequential: SequentialConfig | None = None
    parallel: ParallelConfig | None = None
    environment: EnvironmentConfig | None = Field(
        default_factory=lambda: EnvironmentConfig()
    )

    @model_validator(mode="after")
    def validate_config(self):
        if self.sequential is not None and self.parallel is not None:
            raise ValueError("Cannot use both sequential and parallel options")
        if self.sequential is None and self.parallel is None:
            self.parallel = ParallelConfig()
        return self

    @property
    def strategy(self):
        if self.sequential is not None:
            return "sequential"
        return "parallel"



from ._logger import Logger


class NoopLogger(Logger):
    def debug(self, message: str, *args, **kwargs) -> None:
        pass

    def info(self, message: str, *args, **kwargs) -> None:
        pass

    def success(self, message: str, *args, **kwargs) -> None:
        pass

    def warning(self, message: str, *args, **kwargs) -> None:
        pass

    def error(self, message: str, *args, **kwargs) -> None:
        pass



import abc
from ...component import Component
from ldm.logger import Logger


class Scheme(Component, metaclass=abc.ABCMeta):
    def __init__(
        self,
        name: str,
        *,
        logger: Logger | None = None,
    ) -> None:
        super().__init__(logger)
        self.name = name

    @abc.abstractmethod
    def install(self, entry: str) -> None:
        raise NotImplementedError



from urllib.parse import urlparse, ParseResult
from pydantic import BaseModel
from ._factory import SchemeFactory
from ..scheme import GithubScheme


class GithubSchemeFactoryConfig(BaseModel):
    owner: str
    repo: str
    ref: str | None = None
    token: str | None = None

    @staticmethod
    def from_url(config: dict) -> "GithubSchemeFactoryConfig":
        url = config["url"]
        parsed: ParseResult = urlparse(url)
        path = parsed.path.lstrip("/")
        owner, repo = path.split("/")

        return GithubSchemeFactoryConfig(
            owner=owner,
            repo=repo,
            ref=config.get("ref"),
            token=config.get("token"),
        )


class GithubSchemeFactory(SchemeFactory):
    name = "github"

    def create(
        self,
        config: dict | None = None,
    ) -> GithubScheme:
        config: GithubSchemeFactoryConfig = config or {}
        if "url" in config:
            config = GithubSchemeFactoryConfig.from_url(config)
        else:
            config = GithubSchemeFactoryConfig(**config)
        return GithubScheme(
            name=self.name,
            owner=config.owner,
            repo=config.repo,
            ref=config.ref,
            token=config.token,
            logger=self.logger,
        )



from concurrent.futures import ThreadPoolExecutor
from ...config import ParallelConfig
from .._dependency import Dependency
from ._strategy import InstallStrategy


class ParallelInstallStrategy(InstallStrategy[ParallelConfig]):
    def install(self, dependencies):
        self.logger.debug("Installing dependencies in parallel")
        with ThreadPoolExecutor(max_workers=self.config.workers) as executor:

            def create_install_task(dependency: Dependency) -> None:
                def install():
                    self.logger.debug(f"Installing: {dependency.name}")
                    dependency.install()
                    self.logger.info(f"Installed: {dependency.name}")

                return install

            futures = [
                executor.submit(create_install_task(dependency))
                for dependency in dependencies
            ]

            for future in futures:
                future.result()



import abc
from ldm.logger import Logger, NoopLogger


class Component(metaclass=abc.ABCMeta):
    def __init__(self, logger: Logger | None = None) -> None:
        self.logger = logger or NoopLogger()



from typing import Generator
from urllib.parse import urlparse
from ldm.logger import Logger
from ..config import DependencyConfig, SchemeConfig
from ..scheme import Scheme
from ..factory import HTTPSchemeFactory, HTTPSSchemeFactory, GithubSchemeFactory
from ...component import Component
from ..parse import parse_entry
from ._dependency import Dependency
from .strategy import create_install_strategy


class Installer(Component):
    def __init__(
        self,
        *,
        logger: Logger | None = None,
    ) -> None:
        super().__init__(logger)
        self.schemes: dict[str, Scheme] = {
            # Default schemes
            "http": HTTPSchemeFactory(logger=self.logger).create({}),
            "https": HTTPSSchemeFactory(logger=self.logger).create({}),
        }

    def install(
        self,
        config: DependencyConfig,
        targets: list[str] | None = None,
    ) -> None:
        def parse_schemes(schemes: dict[str, SchemeConfig]) -> dict[str, Scheme]:
            def create_scheme(
                name: str,
                scheme_config: SchemeConfig,
            ) -> tuple[str, Scheme]:
                factory = {
                    "github": GithubSchemeFactory,
                    "http": HTTPSchemeFactory,
                    "https": HTTPSSchemeFactory,
                }[scheme_config.uses]
                return name, factory(logger=self.logger).create(scheme_config.with_)

            def create_schemes(
                schemes: dict[str, SchemeConfig],
            ) -> Generator[tuple[str, Scheme], None, None]:
                for name, scheme_config in schemes.items():
                    self.logger.debug(f"Parsing scheme: {name}")
                    yield create_scheme(name, scheme_config)

            self.logger.debug("Parsing schemes")
            return dict(list(create_schemes(schemes)))

        def parse_dependencies(
            dependencies: dict[str, str],
        ) -> Generator[Dependency, None, None]:
            def decide_scheme(entry: str) -> str:
                source, _ = parse_entry(entry)
                parsed = urlparse(source)
                return parsed.scheme

            def parse_dependency(
                name: str,
                entry: str,
            ) -> Dependency:
                scheme = decide_scheme(entry)
                return Dependency(
                    name,
                    lambda: self.schemes[scheme].install(entry),
                )

            self.logger.debug("Parsing dependencies")
            for name, entry in dependencies.items():
                self.logger.debug(f"Parsing dependency: {name}")
                yield parse_dependency(name, entry)

        self.logger.info("Installing dependencies...")

        self.schemes.update(parse_schemes(config.schemes))

        strategy = create_install_strategy(
            config.config.strategy,
            config.config,
            logger=self.logger,
        )

        dependencies = list(parse_dependencies(config.dependencies))

        if targets is not None:
            self.logger.debug("`targets`_ are given. Filtering dependencies")
            dependencies = list(filter(lambda d: d.name in targets, dependencies))

        strategy.install(dependencies)

        self.logger.success("Dependencies installed")



import abc
from typing import Literal, TypeAlias
from ldm.utils.mapping import Subscriptable

Level: TypeAlias = Literal[
    1,  # DEBUG
    2,  # INFO
    3,  # SUCCESS
    4,  # WARNING
    5,  # ERROR
]


class Logger(Subscriptable, metaclass=abc.ABCMeta):
    def __init__(self, level: Level = 2) -> None:
        self.level = level

    @abc.abstractmethod
    def debug(self, message: str, *args, **kwargs) -> None:
        pass

    @abc.abstractmethod
    def info(self, message: str, *args, **kwargs) -> None:
        pass

    @abc.abstractmethod
    def success(self, message: str, *args, **kwargs) -> None:
        pass

    @abc.abstractmethod
    def warning(self, message: str, *args, **kwargs) -> None:
        pass

    @abc.abstractmethod
    def error(self, message: str, *args, **kwargs) -> None:
        pass



from pydantic import BaseModel
from ..scheme import HTTPScheme
from ._factory import SchemeFactory


class HTTPSchemeFactoryConfig(BaseModel):
    encoding: str | None = None
    headers: dict | None = None
    cookies: dict | None = None


class HTTPSchemeFactory(SchemeFactory):
    name = "http"

    def create(
        self,
        config: dict | None = None,
    ) -> HTTPScheme:
        config: HTTPSchemeFactoryConfig = HTTPSchemeFactoryConfig(**(config or {}))
        return HTTPScheme(
            name=self.name,
            encoding=config.encoding,
            headers=config.headers,
            cookies=config.cookies,
            logger=self.logger,
        )



from ._scheme import Scheme
from ldm.logger import Logger
from urllib.parse import urlparse
from gget import get_remote_file


class GithubScheme(Scheme):
    def __init__(
        self,
        name: str,
        owner: str,
        repo: str,
        ref: str = "HEAD",
        token: str | None = None,
        *,
        logger: Logger | None = None,
    ) -> None:
        super().__init__(name, logger=logger)
        self.owner = owner
        self.repo = repo
        self.ref = ref
        self.token = token

    def install(self, entry: str) -> None:
        # Parse the entry
        source, destination = (
            entry.replace(" ", "").replace("\t", "").replace("\n", "").split("->")
        )
        parsed = urlparse(source)

        path = parsed.path

        get_remote_file(
            path=path,
            reponame=self.repo,
            owner=self.owner,
            output=destination,
            ref=self.ref,
            provider="github",
            provider_options={"token": self.token},
        )



class Subscriptable:
    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    def get(self, key, default=None):
        return getattr(self, key, default)

    def __contains__(self, key):
        return hasattr(self, key)

    def __iter__(self):
        return iter(self.__dict__.keys())

    def __len__(self):
        return len(self.__dict__)

    def items(self):
        return self.__dict__.items()

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()



from typing import Literal, Any
from ldm.logger import Logger
from ._strategy import InstallStrategy
from ._sequential import SequentialInstallStrategy
from ._parallel import ParallelInstallStrategy


def create_install_strategy(
    strategy: Literal["sequential", "parallel"],
    config: Any,
    logger: Logger | None = None,
) -> InstallStrategy:
    logger.debug(f"Creating install strategy: {strategy}")
    if strategy == "sequential":
        return SequentialInstallStrategy(config.sequential, logger=logger)
    elif strategy == "parallel":
        return ParallelInstallStrategy(config.parallel, logger=logger)
    else:
        raise ValueError(f"Unknown strategy: {strategy}")



# flake8: noqa: F401

from ._create import create_install_strategy
from ._strategy import InstallStrategy
from ._sequential import SequentialInstallStrategy
from ._parallel import ParallelInstallStrategy



from typing import Callable
from ..logger import Logger, ClickLogger


def create_logger(debug: bool) -> ClickLogger:
    level = 1 if debug else 2
    return ClickLogger(level=level)


def error_boundary(func: Callable, logger: Logger):
    try:
        func()
    except Exception as e:
        for line in str(e).split("\n"):
            logger.error(line)
        exit(1)
